package ru.dbotthepony.kstarbound.util

import ru.dbotthepony.kstarbound.io.json.EitherTypeAdapter

/**
 * Представляет собой контейнер с "или тот или другой" значениями
 *
 * JSON адаптер реализуется через [EitherTypeAdapter]
 */
sealed class Either<L, R> {
	class Left<L, R>(val value: L) : Either<L, R>() {
		override val isLeft: Boolean
			get() = true
		override val isRight: Boolean
			get() = false

		override fun <T> map(left: (L) -> T, right: (R) -> T): T {
			return left.invoke(this.value)
		}

		override fun left(): L {
			return value
		}

		override fun right(): R {
			throw NoSuchElementException()
		}
	}

	class Right<L, R>(val value: R) : Either<L, R>() {
		override val isLeft: Boolean
			get() = false
		override val isRight: Boolean
			get() = true

		override fun <T> map(left: (L) -> T, right: (R) -> T): T {
			return right.invoke(value)
		}

		override fun left(): L {
			throw NoSuchElementException()
		}

		override fun right(): R {
			return value
		}
	}

	abstract val isLeft: Boolean
	abstract val isRight: Boolean

	abstract fun <T> map(left: (L) -> T, right: (R) -> T): T

	/**
	 * @throws NoSuchElementException
	 */
	abstract fun left(): L

	/**
	 * @throws NoSuchElementException
	 */
	abstract fun right(): R

	inline fun leftOrElse(orElse: () -> L): L {
		if (isLeft)
			return left()
		else
			return orElse.invoke()
	}

	inline fun rightOrElse(orElse: () -> R): R {
		if (isRight)
			return right()
		else
			return orElse.invoke()
	}

	companion object {
		fun <L, R> left(value: L): Either<L, R> {
			return Left(value)
		}

		fun <L, R> right(value: R): Either<L, R> {
			return Right(value)
		}
	}
}
